import json
from collections import defaultdict

import requests
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import permission_required
from django.contrib.contenttypes.models import ContentType
from django.http import Http404, HttpResponse, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from rest_framework import status
from rest_framework.fields import ReadOnlyField
from wagtail.models import Page

from .auth import check_digest, digest_for_source
from .locators import get_locator_for_model
from .models import get_model_for_path
from .operations import ImportPlanner
from .serializers import serializer_registry
from .vendor.wagtail_admin_api.serializers import AdminPageSerializer
from .vendor.wagtail_admin_api.views import PagesAdminAPIViewSet


def pages_for_export(request, root_page_id):
    check_digest(str(root_page_id), request.GET.get('digest', ''))

    root_page = get_object_or_404(Page, id=root_page_id)

    pages = [root_page.specific] if request.GET.get('recursive', 'true') == 'false' else root_page.get_descendants(inclusive=True).specific()

    ids_for_import = [
        ['wagtailcore.page', page.pk] for page in pages
    ]

    objects = []
    object_references = set()

    models_to_serialize = set(pages)
    serialized_models = set()

    while models_to_serialize:
        model = models_to_serialize.pop()
        serializer = serializer_registry.get_model_serializer(type(model))
        objects.append(serializer.serialize(model))
        object_references.update(serializer.get_object_references(model))
        models_to_serialize.update(serializer.get_objects_to_serialize(model).difference(serialized_models))

    mappings = []
    for model, pk in object_references:
        uid = get_locator_for_model(model).get_uid_for_local_id(pk)
        mappings.append(
            [model._meta.label_lower, pk, uid]
        )

    return JsonResponse({
        'ids_for_import': ids_for_import,
        'mappings': mappings,
        'objects': objects,
    }, json_dumps_params={'indent': 2})


def models_for_export(request, model_path, object_id=None):
    """
    Return data for a specific model based on the incoming model_path.

    If an object_id is provided, search for a single model object.
    """
    check_digest(str(model_path), request.GET.get('digest', ''))

    # 1. Confirm whether or not th model_path leads to a real model.
    app_label, model_name = model_path.split('.')
    Model = ContentType.objects.get_by_natural_key(app_label, model_name).model_class()

    if object_id is None:
        model_objects = Model.objects.all()
    else:
        model_objects = [Model.objects.get(pk=object_id)]

    # 2. If this was just a model and not a specific object, get all child IDs.
    ids_for_import = [
        [model_path, obj.pk] for obj in model_objects
    ]

    objects = []
    object_references = set()

    models_to_serialize = set(model_objects)
    serialized_models = set()

    while models_to_serialize:
        model = models_to_serialize.pop()
        serializer = serializer_registry.get_model_serializer(type(model))
        objects.append(serializer.serialize(model))
        object_references.update(serializer.get_object_references(model))
        models_to_serialize.update(serializer.get_objects_to_serialize(model).difference(serialized_models))

    mappings = []
    for model, pk in object_references:
        uid = get_locator_for_model(model).get_uid_for_local_id(pk)
        mappings.append(
            [model._meta.label_lower, pk, uid]
        )

    return JsonResponse({
        'ids_for_import': ids_for_import,
        'mappings': mappings,
        'objects': objects,
    }, json_dumps_params={'indent': 2})


@csrf_exempt
@require_POST
def objects_for_export(request):
    """
    Accepts a POST request with a JSON payload structured as:
        {
            'model_label': [list of IDs],
            'model_label': [list of IDs],
        }
    and returns an API response with objects / mappings populated (but ids_for_import empty).
    """

    check_digest(request.body, request.GET.get('digest', ''))

    request_data = json.loads(request.body.decode('utf-8'))

    objects = []
    object_references = set()
    serialized_models = set()
    models_to_serialize = set()

    for model_path, ids in request_data.items():
        model = get_model_for_path(model_path)
        serializer = serializer_registry.get_model_serializer(model)

        models_to_serialize.update(serializer.get_objects_by_ids(ids))
        while models_to_serialize:
            instance = models_to_serialize.pop()
            serializer = serializer_registry.get_model_serializer(type(instance))
            objects.append(serializer.serialize(instance))
            object_references.update(serializer.get_object_references(instance))
            models_to_serialize.update(serializer.get_objects_to_serialize(instance).difference(serialized_models))

    mappings = []
    for model, pk in object_references:
        uid = get_locator_for_model(model).get_uid_for_local_id(pk)
        mappings.append(
            [model._meta.label_lower, pk, uid]
        )

    return JsonResponse({
        'ids_for_import': [],
        'mappings': mappings,
        'objects': objects,
    }, json_dumps_params={'indent': 2})


class UIDField(ReadOnlyField):
    """
    Serializes UID for the Page Chooser API
    """
    def get_attribute(self, instance):
        return get_locator_for_model(Page).get_uid_for_local_id(instance.id, create=False)


class TransferPageChooserSerializer(AdminPageSerializer):
    uid = UIDField(read_only=True)


class PageChooserAPIViewSet(PagesAdminAPIViewSet):
    base_serializer_class = TransferPageChooserSerializer
    meta_fields = PagesAdminAPIViewSet.meta_fields + [
        'uid'
    ]
    listing_default_fields = PagesAdminAPIViewSet.listing_default_fields + [
        'uid'
    ]


@permission_required(
    "wagtail_transfer.wagtailtransfer_can_import", login_url="wagtailadmin_login"
)
def chooser_api_proxy(request, source_name, path):
    source_config = getattr(settings, 'WAGTAILTRANSFER_SOURCES', {}).get(source_name)

    api_proxy_timeout_seconds = getattr(settings, 'WAGTAILTRANSFER_CHOOSER_API_PROXY_TIMEOUT', 5)

    if source_config is None:
        raise Http404("Source does not exist")

    default_chooser_endpoint = 'pages'
    if 'models' in request.GET:
        default_chooser_endpoint = 'models'

    base_url = source_config['BASE_URL'] + 'api/chooser/{}/'.format(default_chooser_endpoint)

    message = request.GET.urlencode()
    digest = digest_for_source(source_name, message)

    response = requests.get(f"{base_url}{path}?{message}&digest={digest}", headers={
        'Accept': request.META['HTTP_ACCEPT'],
    }, timeout=api_proxy_timeout_seconds)

    return HttpResponse(response.content, status=response.status_code)


@permission_required(
    "wagtail_transfer.wagtailtransfer_can_import", login_url="wagtailadmin_login"
)
def choose_page(request):
    return render(request, 'wagtail_transfer/choose_page.html', {
        'sources_data': json.dumps([
            {
                'value': source_name,
                'label': source_name,
                'page_chooser_api': reverse('wagtail_transfer_admin:chooser_api_proxy', args=[source_name, ''])
            }
            for source_name in getattr(settings, 'WAGTAILTRANSFER_SOURCES', {}).keys()
        ]),
    })


def import_missing_object_data(source, importer: ImportPlanner):
    base_url = settings.WAGTAILTRANSFER_SOURCES[source]['BASE_URL']
    while importer.missing_object_data:
        # convert missing_object_data from a set of (model_class, id) tuples
        # into a dict of {model_class_label: [list_of_ids]}
        missing_object_data_by_type = defaultdict(list)
        for model_class, source_id in importer.missing_object_data:
            missing_object_data_by_type[model_class].append(source_id)

        request_data = json.dumps({
            model_class._meta.label_lower: ids
            for model_class, ids in missing_object_data_by_type.items()
        })
        digest = digest_for_source(source, request_data)

        # request the missing object data and add to the import plan
        response = requests.post(
            f"{base_url}api/objects/", params={'digest': digest}, data=request_data
        )
        importer.add_json(response.content)
    importer.run()
    return importer


def import_page(request):
    source = request.POST['source']
    base_url = settings.WAGTAILTRANSFER_SOURCES[source]['BASE_URL']
    digest = digest_for_source(source, str(request.POST['source_page_id']))

    response = requests.get(f"{base_url}api/pages/{request.POST['source_page_id']}/", params={'digest': digest})

    dest_page_id = request.POST['dest_page_id'] or None
    importer = ImportPlanner.for_page(source=request.POST['source_page_id'], destination=dest_page_id)
    importer.add_json(response.content)
    importer = import_missing_object_data(source, importer)

    if dest_page_id:
        return redirect('wagtailadmin_explore', dest_page_id)
    else:
        return redirect('wagtailadmin_explore_root')


def import_model(request):
    source = request.POST['source']
    model = request.POST['source_model']
    base_url = settings.WAGTAILTRANSFER_SOURCES[source]['BASE_URL']
    digest = digest_for_source(source, model)

    url = f"{base_url}api/models/{model}/"
    if request.POST.get("source_model_object_id"):
        source_model_object_id = request.POST.get("source_model_object_id")
        url = f"{url}{source_model_object_id}/"

    response = requests.get(url, params={'digest': digest})
    importer = ImportPlanner.for_model(model=model)
    importer.add_json(response.content)
    importer = import_missing_object_data(source, importer)

    messages.add_message(request, messages.SUCCESS, 'Snippet(s) successfully imported')
    app_label, model_name = model.split('.')
    return redirect(f'wagtailsnippets_{app_label}_{model_name}:list')


@permission_required(
    "wagtail_transfer.wagtailtransfer_can_import", login_url="wagtailadmin_login"
)
@require_POST
def do_import(request):
    post_type = request.POST.get('type', 'page')
    if post_type == 'page':
        return import_page(request)
    elif post_type == 'model':
        return import_model(request)


def check_page_existence_for_uid(request):
    """
    Check whether a page with the specified UID exists - used for checking whether a page has already been imported
    to the destination site
    """
    uid = request.GET.get('uid', '')
    locator = get_locator_for_model(Page)
    page_exists = bool(locator.find(uid))
    result = status.HTTP_200_OK if page_exists else status.HTTP_404_NOT_FOUND
    return HttpResponse('', status=result)



import functools

from django.urls import include, re_path
from wagtail.utils.urlpatterns import decorate_urlpatterns


class WagtailAPIRouter:
    """
    A class that provides routing and cross-linking for a collection
    of API endpoints
    """
    def __init__(self, url_namespace):
        self.url_namespace = url_namespace
        self._endpoints = {}

    def register_endpoint(self, name, class_):
        self._endpoints[name] = class_

    def get_model_endpoint(self, model):
        """
        Finds the endpoint in the API that represents a model

        Returns a (name, endpoint_class) tuple. Or None if an
        endpoint is not found.
        """
        for name, class_ in self._endpoints.items():
            if issubclass(model, class_.model):
                return name, class_

    def get_model_listing_urlpath(self, model):
        """
        Returns a URL path (excluding scheme and hostname) to the listing
        page of a model

        Returns None if the model is not represented by any endpoints.
        """
        endpoint = self.get_model_endpoint(model)

        if endpoint:
            endpoint_name, endpoint_class = endpoint[0], endpoint[1]
            url_namespace = self.url_namespace + ':' + endpoint_name
            return endpoint_class.get_model_listing_urlpath(model, namespace=url_namespace)

    def get_object_detail_urlpath(self, model, pk):
        """
        Returns a URL path (excluding scheme and hostname) to the detail
        page of an object.

        Returns None if the object is not represented by any endpoints.
        """
        endpoint = self.get_model_endpoint(model)

        if endpoint:
            endpoint_name, endpoint_class = endpoint[0], endpoint[1]
            url_namespace = self.url_namespace + ':' + endpoint_name
            return endpoint_class.get_object_detail_urlpath(model, pk, namespace=url_namespace)

    def wrap_view(self, func):
        @functools.wraps(func)
        def wrapped(request, *args, **kwargs):
            request.wagtailapi_router = self
            return func(request, *args, **kwargs)

        return wrapped

    def get_urlpatterns(self):
        urlpatterns = []

        for name, class_ in self._endpoints.items():
            pattern = re_path(
                r'^{}/'.format(name),
                include((class_.get_urlpatterns(), name), namespace=name)
            )
            urlpatterns.append(pattern)

        decorate_urlpatterns(urlpatterns, self.wrap_view)

        return urlpatterns

    @property
    def urls(self):
        """
        A shortcut to allow quick registration of the API in a URLconf.

        Use with Django's include() function:

            re_path(r'api/', include(myapi.urls)),
        """
        return self.get_urlpatterns(), self.url_namespace, self.url_namespace



import logging
import json
import pathlib
from functools import lru_cache
from urllib.parse import urlparse

from django.conf import settings
from django.contrib.contenttypes.fields import (GenericForeignKey,
                                                GenericRelation)
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models.fields.reverse_related import ManyToOneRel
from django.utils.encoding import is_protected_type
from django.utils.functional import cached_property
from modelcluster.fields import ParentalKey
from taggit.managers import TaggableManager
from wagtail import hooks
from wagtail.fields import RichTextField, StreamField

from .files import File, FileTransferError, get_file_hash, get_file_size
from .locators import get_locator_for_model
from .models import get_base_model, get_base_model_for_path
from .richtext import get_reference_handler
from .streamfield import get_object_references, update_object_ids


logger = logging.getLogger(__name__)

WAGTAILTRANSFER_FOLLOWED_REVERSE_RELATIONS = getattr(settings, "WAGTAILTRANSFER_FOLLOWED_REVERSE_RELATIONS", [('wagtailimages.image', 'tagged_items', True)])
FOLLOWED_REVERSE_RELATIONS = {
    (model_label.lower(), relation.lower()) for model_label, relation, _ in WAGTAILTRANSFER_FOLLOWED_REVERSE_RELATIONS
}
DELETED_REVERSE_RELATIONS = {
    (model_label.lower(), relation.lower()) for model_label, relation, track_deletions in WAGTAILTRANSFER_FOLLOWED_REVERSE_RELATIONS if track_deletions
}
ADMIN_BASE_URL = getattr(
    settings, "WAGTAILADMIN_BASE_URL",
    getattr(settings, "BASE_URL", None)
)


class FieldAdapter:
    def __init__(self, field):
        self.field = field
        self.name = self.field.name

    def serialize(self, instance):
        """
        Retrieve the value for this field from the given model instance, and return a
        representation of it that can be included in JSON data
        """
        value = self.field.value_from_object(instance)
        if not is_protected_type(value):
            value = self.field.value_to_string(instance)

        return value

    def get_object_references(self, instance):
        """
        Return a set of (model_class, id) pairs for all objects referenced in this field
        """
        return set()

    def get_dependencies(self, value):
        """
        A set of (base_model_class, id, is_hard) tuples for objects that must exist at the
        destination before populate_field can proceed with the given value.

        is_hard is a boolean - if True, then the object MUST exist in order for populate_field to
            succeed; if False, then the operation can still complete without it (albeit possibly
            with broken links).

        This differs from get_object_references in that references inside related child objects
        do not need to be considered, as they do not block the creation/update of the parent
        object.
        """
        return set()

    def get_object_deletions(self, instance, value, context):
        """
        A set of (base_model_class, id) tuples for objects that must be deleted at the destination site
        """
        return set()

    def update_object_references(self, value, destination_ids_by_source):
        """
        Return a modified version of value with object references replaced by their corresponding
        entries in destination_ids_by_source - a mapping of (model_class, source_id) to
        destination_id
        """
        return value

    def populate_field(self, instance, value, context):
        """
        Populate this field on the passed model instance, given a value in its serialized form
        as returned by `serialize`
        """
        value = self.update_object_references(value, context.destination_ids_by_source)
        setattr(instance, self.field.get_attname(), self.field.to_python(value))

    def get_managed_fields(self):
        """
        Normally, a FieldAdapter will adapt a single field. However, more complex fields like
        GenericForeignKey may 'manage' several other fields. get_managed_fields returns a list of names
        of managed fields, whose field adapters should not be used when serializing the model. Note
        that if a managed field also has managed fields itself, these will also be ignored when
        serializing the model - the current field adapter is expected to address all managed fields in
        the chain.
        """
        return []

    def get_objects_to_serialize(self, instance):
        """
        Return a set of (model_class, id) pairs for objects that should be serialized on export, before
        it is known whether or not they exist or should be updated at the destination site
        """
        return set()



class ForeignKeyAdapter(FieldAdapter):
    def __init__(self, field):
        super().__init__(field)
        self.related_base_model = get_base_model(self.field.related_model)

    def get_object_references(self, instance):
        pk = self.field.value_from_object(instance)
        if pk is None:
            return set()
        else:
            return {(self.related_base_model, pk)}

    def get_dependencies(self, value):
        if value is None:
            return set()
        elif self.field.blank and self.field.null:
            # field is nullable, so it's a soft dependency; we can leave the field empty in the
            # case that the target object cannot be created
            return {(self.related_base_model, value, False)}
        else:
            # this is a hard dependency
            return {(self.related_base_model, value, True)}

    def update_object_references(self, value, destination_ids_by_source):
        return destination_ids_by_source.get((self.related_base_model, value))


class GenericForeignKeyAdapter(FieldAdapter):
    def serialize(self, instance):
        linked_instance = getattr(instance, self.field.name, None)
        if linked_instance:
            # here we do not use the base model, as the GFK could be pointing specifically at the child
            # which needs to be represented accurately
            return (linked_instance._meta.label_lower, linked_instance.pk)

    def get_object_references(self, instance):
        linked_instance = getattr(instance, self.field.name, None)
        if linked_instance:
            return {(get_base_model(linked_instance), linked_instance.pk)}
        return set()

    def get_dependencies(self, value):
        if value is None:
            return set()

        model_path, model_id = value
        base_model = get_base_model_for_path(model_path)

        # GenericForeignKey itself has no blank or null properties, so we need to determine its nullable status
        # from the underlying fields it uses
        options = self.field.model._meta
        ct_field = options.get_field(self.field.ct_field)
        fk_field = options.get_field(self.field.ct_field)

        if all((ct_field.blank, ct_field.null, fk_field.blank, fk_field.null)):
            # field is nullable, so it's a soft dependency; we can leave the field empty in the
            # case that the target object cannot be created
            return {(base_model, model_id, False)}
        else:
            # this is a hard dependency
            return {(base_model, model_id, True)}

    def update_object_references(self, value, destination_ids_by_source):
        if value:
            model_path, model_id = value
            base_model = get_base_model_for_path(model_path)
            return (model_path, destination_ids_by_source.get((base_model, model_id)))

    def populate_field(self, instance, value, context):
        model_id, content_type = None, None
        if value:
            model_path, model_id = self.update_object_references(value, context.destination_ids_by_source)
            content_type = ContentType.objects.get_by_natural_key(*model_path.split('.'))

        setattr(instance, instance._meta.get_field(self.field.ct_field).get_attname(), content_type.pk)
        setattr(instance, self.field.fk_field, model_id)

    def get_managed_fields(self):
        return [self.field.fk_field, self.field.ct_field]


class ManyToOneRelAdapter(FieldAdapter):
    def __init__(self, field):
        super().__init__(field)
        self.related_field = getattr(field, 'field', None) or getattr(field, 'remote_field', None)
        self.related_base_model = get_base_model(field.related_model)
        self.is_parental = isinstance(self.related_field, ParentalKey)
        self.is_followed = (get_base_model(self.field.model)._meta.label_lower, self.name) in FOLLOWED_REVERSE_RELATIONS
        if self.is_parental or self.is_followed:
            # presumably this info is most useful when it's going to be used, so this avoids some extra log noise
            logger.debug("Field adaptor registered: "
                         f"{self.field}, {get_base_model(self.field.model)._meta.label_lower, self.name})")

    def _get_related_objects(self, instance):
        results = getattr(instance, self.name).all()
        if results:
            logger.debug("Related objects found for "
                         f"{self.name}, {instance}: {results}")
        else:
            logger.debug(f"No related objects found for {self.name}")
        return results

    def serialize(self, instance):
        if self.is_parental or self.is_followed:
            return list(self._get_related_objects(instance).values_list('pk', flat=True))

    def get_object_references(self, instance):
        refs = set()
        if self.is_parental or self.is_followed:
            for pk in self._get_related_objects(instance).values_list('pk', flat=True):
                refs.add((self.related_base_model, pk))
        else:
            logger.debug(f"{self.field}, {get_base_model(self.field.model)._meta.label_lower, self.name}"
                         " is not parental or followed, not adding to refs")
        return refs

    def get_object_deletions(self, instance, value, context):
        if (self.is_parental or (get_base_model(self.field.model)._meta.label_lower, self.name) in DELETED_REVERSE_RELATIONS):
            value = value or []
            uids = {context.uids_by_source[(self.related_base_model, pk)] for pk in value}
            # delete any related objects on the existing object if they can't be mapped back
            # to one of the uids in the new set
            locator = get_locator_for_model(self.related_base_model)
            matched_destination_ids = set()
            for uid in uids:
                child = locator.find(uid)
                if child is not None:
                    matched_destination_ids.add(child.pk)
            logger.debug(f"Deleting the following objects: {matched_destination_ids}")
            return {child for child in self._get_related_objects(instance) if child.pk not in matched_destination_ids}
        return set()

    def get_objects_to_serialize(self, instance):
        if self.is_parental:
            return getattr(instance, self.name).all()
        return set()

    def populate_field(self, instance, value, context):
        pass


class RichTextAdapter(FieldAdapter):
    def get_object_references(self, instance):
        return get_reference_handler().get_objects(self.field.value_from_object(instance))

    def get_dependencies(self, value):
        return {
            (model, id, False)  # references in rich text are soft dependencies
            for model, id in get_reference_handler().get_objects(value)
        }

    def update_object_references(self, value, destination_ids_by_source):
        return get_reference_handler().update_ids(value, destination_ids_by_source)


class StreamFieldAdapter(FieldAdapter):
    def __init__(self, field):
        super().__init__(field)
        self.stream_block = self.field.stream_block

    def get_object_references(self, instance):
        # get the list of dicts representation of the streamfield json
        stream = self.stream_block.get_prep_value(self.field.value_from_object(instance))
        return get_object_references(self.stream_block, stream)

    def get_dependencies(self, value):
        return {
            (model, id, False)  # references in streamfield are soft dependencies
            for model, id in get_object_references(self.stream_block, json.loads(value))
        }

    def update_object_references(self, value, destination_ids_by_source):
        return json.dumps(update_object_ids(self.stream_block, json.loads(value), destination_ids_by_source))


class FileAdapter(FieldAdapter):
    def serialize(self, instance):
        value = self.field.value_from_object(instance)
        if not value:
            return None
        url = value.url
        if url.startswith('/'):
            # Using a relative media url. ie. /media/
            # Prepend the BASE_URL to turn this into an absolute URL
            if ADMIN_BASE_URL is None:
                raise ImproperlyConfigured(
                    "A WAGTAILADMIN_BASE_URL or BASE_URL setting must be provided when importing files"
                )
            url = ADMIN_BASE_URL.rstrip('/') + url
        return {
            'download_url': url,
            'size': get_file_size(self.field, instance),
            'hash': get_file_hash(self.field, instance),
        }

    def populate_field(self, instance, value, context):
        if not value:
            return None
        imported_file = context.imported_files_by_source_url.get(value['download_url'])
        if imported_file is None:

            existing_file = self.field.value_from_object(instance)

            if existing_file:
                existing_file_hash = get_file_hash(self.field, instance)
                if existing_file_hash == value['hash']:
                    # File not changed, so don't bother updating it
                    return

            # Get the local filename
            name = pathlib.PurePosixPath(urlparse(value['download_url']).path).name
            local_filename = self.field.generate_filename(instance, name)

            _file = File(local_filename, value['size'], value['hash'], value['download_url'])
            try:
                imported_file = _file.transfer()
            except FileTransferError:
                return None
            context.imported_files_by_source_url[_file.source_url] = imported_file

        value = imported_file.file.name
        getattr(instance, self.field.get_attname()).name = value


class ManyToManyFieldAdapter(FieldAdapter):
    def __init__(self, field):
        super().__init__(field)
        self.related_base_model = get_base_model(self.field.related_model)

    def _get_pks(self, instance):
        return [model.pk for model in self.field.value_from_object(instance)]

    def get_object_references(self, instance):
        refs = set()
        for pk in self._get_pks(instance):
            refs.add((self.related_base_model, pk))
        return refs

    def get_dependencies(self, value):
        return {(self.related_base_model, id, False) for id in value}

    def serialize(self, instance):
        pks = list(self._get_pks(instance))
        return pks

    def populate_field(self, instance, value, context):
        # setting forward ManyToMany directly is prohibited
        pass


class TaggableManagerAdapter(FieldAdapter):
    def populate_field(self, instance, value, context):
        # TODO
        pass


class GenericRelationAdapter(ManyToOneRelAdapter):
    pass


class AdapterRegistry:
    BASE_ADAPTERS_BY_FIELD_CLASS = {
        models.Field: FieldAdapter,
        models.ForeignKey: ForeignKeyAdapter,
        ManyToOneRel: ManyToOneRelAdapter,
        RichTextField: RichTextAdapter,
        StreamField: StreamFieldAdapter,
        models.FileField: FileAdapter,
        models.ManyToManyField: ManyToManyFieldAdapter,
        TaggableManager: TaggableManagerAdapter,
        GenericRelation: GenericRelationAdapter,
        GenericForeignKey: GenericForeignKeyAdapter,
    }

    def __init__(self):
        self._scanned_for_adapters = False
        self.adapters_by_field_class = {}

    def _scan_for_adapters(self):
        adapters = dict(self.BASE_ADAPTERS_BY_FIELD_CLASS)

        for fn in hooks.get_hooks('register_field_adapters'):
            adapters.update(fn())

        self.adapters_by_field_class = adapters
        self._scanned_for_adapters = True

    @lru_cache(maxsize=None)
    def get_field_adapter(self, field):
        # find the adapter class for the most specific class in the field's inheritance tree

        if not self._scanned_for_adapters:
            self._scan_for_adapters()

        for field_class in type(field).__mro__:
            if field_class in self.adapters_by_field_class:
                adapter_class = self.adapters_by_field_class[field_class]
                return adapter_class(field)


adapter_registry = AdapterRegistry()



from functools import lru_cache

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from treebeard.mp_tree import MP_Node
from wagtail import hooks
from wagtail.models import Page

from .field_adapters import adapter_registry
from .models import get_base_model


def _get_subclasses_recurse(model):
    """
    Given a Model class, find all related objects, exploring children
    recursively, returning a `list` of strings representing the
    relations for select_related, adapted from https://github.com/jazzband/django-model-utils/blob/master/model_utils/managers.py
    """

    related_objects = [f for f in model._meta.get_fields() if isinstance(f, models.OneToOneRel)]

    rels = [
        rel for rel in related_objects
        if isinstance(rel.field, models.OneToOneField)
        and issubclass(rel.field.model, model)
        and model is not rel.field.model
        and rel.parent_link
    ]

    subclasses = []
    for rel in rels:
        for subclass in _get_subclasses_recurse(rel.field.model):
            subclasses.append(
                rel.get_accessor_name() + LOOKUP_SEP + subclass)
        subclasses.append(rel.get_accessor_name())
    return subclasses


def _get_sub_obj_recurse(obj, s):
    """
    Given an object and its potential subclasses in lookup string form,
    retrieve its most specific subclass recursively
    Taken from: https://github.com/jazzband/django-model-utils/blob/master/model_utils/managers.py
    """
    rel, _, s = s.partition(LOOKUP_SEP)

    try:
        node = getattr(obj, rel)
    except ObjectDoesNotExist:
        return None
    if s:
        child = _get_sub_obj_recurse(node, s)
        return child
    else:
        return node


def get_subclass_instances(instances, subclasses):
    subclass_instances = []
    for obj in instances:
        sub_obj = None
        for s in subclasses:
            sub_obj = _get_sub_obj_recurse(obj, s)
            if sub_obj:
                break
        if not sub_obj:
            sub_obj = obj
        subclass_instances.append(sub_obj)
    return subclass_instances


class ModelSerializer:
    ignored_fields = []

    def __init__(self, model):
        self.model = model
        self.base_model = get_base_model(model)

        field_adapters = []
        adapter_managed_fields = []
        for field in self.model._meta.get_fields():
            if field.name in self.ignored_fields:
                continue

            # ignore primary keys (including MTI parent pointers)
            if getattr(field, 'primary_key', False):
                continue

            adapter = adapter_registry.get_field_adapter(field)

            if adapter:
                adapter_managed_fields = adapter_managed_fields + adapter.get_managed_fields()
                field_adapters.append(adapter)

        self.field_adapters = [adapter for adapter in field_adapters if adapter.name not in adapter_managed_fields]

    def get_objects_by_ids(self, ids):
        """
        Given a list of IDs, return a list of model instances that we can
        run serialize and get_object_references on, fetching the specific subclasses
        if using multi table inheritance as appropriate
        """
        base_queryset = self.model.objects.filter(pk__in=ids)
        subclasses = _get_subclasses_recurse(self.model)
        return get_subclass_instances(base_queryset, subclasses)

    def serialize_fields(self, instance):
        return {
            field_adapter.name: field_adapter.serialize(instance)
            for field_adapter in self.field_adapters
        }

    def serialize(self, instance):
        return {
            'model': self.model._meta.label_lower,
            'pk': instance.pk,
            'fields': self.serialize_fields(instance)
        }

    def get_object_references(self, instance):
        refs = {
            # always include the primary key as an object reference
            (self.base_model, instance.pk)
        }
        for f in self.field_adapters:
            refs.update(f.get_object_references(instance))
        return refs

    def get_objects_to_serialize(self, instance):
        objects = set()
        for f in self.field_adapters:
            objects.update(f.get_objects_to_serialize(instance))
        return objects


class TreeModelSerializer(ModelSerializer):
    ignored_fields = ['path', 'depth', 'numchild']

    def serialize(self, instance):
        result = super().serialize(instance)
        if instance.is_root():
            result['parent_id'] = None
        else:
            result['parent_id'] = instance.get_parent().pk

        return result

    def get_object_references(self, instance):
        refs = super().get_object_references(instance)
        if not instance.is_root():
            # add a reference for the parent ID
            refs.add(
                (self.base_model, instance.get_parent().pk)
            )
        return refs


class PageSerializer(TreeModelSerializer):
    ignored_fields = TreeModelSerializer.ignored_fields + [
        'url_path', 'content_type', 'draft_title', 'has_unpublished_changes', 'owner',
        'go_live_at', 'expire_at', 'expired', 'locked', 'first_published_at', 'last_published_at',
        'latest_revision_created_at', 'live_revision',
    ]

    def get_objects_by_ids(self, ids):
        # serialize method needs the instance in its specific form
        return self.model.objects.filter(pk__in=ids).specific()


class SerializerRegistry:
    BASE_SERIALIZERS_BY_MODEL_CLASS = {
        models.Model: ModelSerializer,
        MP_Node: TreeModelSerializer,
        Page: PageSerializer,
    }

    def __init__(self):
        self._scanned_for_serializers = False
        self.serializers_by_model_class = {}

    def _scan_for_serializers(self):
        serializers = dict(self.BASE_SERIALIZERS_BY_MODEL_CLASS)

        for fn in hooks.get_hooks('register_custom_serializers'):
            serializers.update(fn())

        self.serializers_by_model_class = serializers
        self._scanned_for_serializers = True

    @lru_cache(maxsize=None)
    def get_model_serializer(self, model):
        # find the serializer class for the most specific class in the model's inheritance tree

        if not self._scanned_for_serializers:
            self._scan_for_serializers()

        for cls in model.__mro__:
            if cls in self.serializers_by_model_class:
                serializer_class = self.serializers_by_model_class[cls]
                return serializer_class(model)


serializer_registry = SerializerRegistry()



import re
from functools import partial

from wagtail.rich_text import features
from wagtail.rich_text.rewriters import extract_attrs

from .models import get_base_model

FIND_A_TAG = re.compile(r'<a(\b[^>]*)>(.*?)</a>')
FIND_EMBED_TAG = re.compile(r'<embed(\b[^>]*)/>')
FIND_ID = re.compile(r'id="([^"]*)"')


class RichTextReferenceHandler:
    """
    Handles updating ids and retrieving object references for <tag type_attribute="foo" id="my_id" /> tags representing references within rich text. Tags are found using a regex tag_matcher, and their
    specific handler (eg PageLinkHandler) is looked up using a dict, handlers, which maps the value of type_attribute (where type_attribute might be emebedtype)
    to a specific handler.

    The tag matcher must be a compiled regular expression where the first matching group is the tag's body (ie its attributes) and the second the tag's inner contents (if any).
    Note this only works for tags which cannot be nested inside the same tag, so this works fine for eg matching <a> tags since nested <a> tags are illegal.
    """
    def __init__(self, handlers, tag_matcher, type_attribute, destination_ids_by_source={}):
        self.handlers = handlers
        self.tag_matcher = tag_matcher
        self.type_attribute = type_attribute
        self.destination_ids_by_source = destination_ids_by_source

    def update_tag_id(self, match, destination_ids_by_source):
        # Updates a specific tag's id from source to destination Wagtail instance, or removes the tag if no id mapping exists
        tag_body = match.group(1)
        attrs = extract_attrs(tag_body)
        try:
            handler = self.handlers[attrs[self.type_attribute]]
            target_model = get_base_model(handler.get_model())
            new_id = destination_ids_by_source.get((target_model, int(attrs['id'])))
            if new_id is None:
                # Return the tag's inner contents, effectively removing the tag
                try:
                    return match.group(2)
                except IndexError:
                    # The tag has no inner content, return a blank string instead
                    return ''
            # Otherwise update the id and construct the new tag string
            new_tag_body = FIND_ID.sub('id="{0}"'.format(str(new_id)), tag_body)
            tag_body_offset = match.start(0)
            new_tag_string = match.group(0)[:(match.start(1)-tag_body_offset)] + new_tag_body + match.group(0)[(match.end(1)-tag_body_offset):]
            return new_tag_string
        except KeyError:
            # If the relevant handler cannot be found, don't update the tag id
            pass
        return match.group(0)

    def get_objects(self, html):
        # Gets object references
        objects = set()
        if html:
            for match in self.tag_matcher.finditer(html):
                attrs = extract_attrs(match.group(1))
                try:
                    handler = self.handlers[attrs[self.type_attribute]]
                    objects.add((get_base_model(handler.get_model()), int(attrs['id'])))
                except KeyError:
                    # If no handler can be found, no object reference can be added.
                    # This might occur when the link is a plain url
                    pass
        return objects

    def update_ids(self, html, destination_ids_by_source):
        # Update source instance ids to destination instance ids when possible
        if html is None:
            return None
        else:
            return self.tag_matcher.sub(partial(self.update_tag_id, destination_ids_by_source=destination_ids_by_source), html)


class MultiTypeRichTextReferenceHandler:
    """Handles retrieving object references and updating ids for several different kinds of tags in rich text"""
    def __init__(self, handlers):
        self.handlers = handlers

    def update_ids(self, html, destination_ids_by_source):
        for handler in self.handlers:
            html = handler.update_ids(html, destination_ids_by_source)
        return html

    def get_objects(self, html):
        objects = set()
        for handler in self.handlers:
            objects = objects.union(handler.get_objects(html))
        return objects


REFERENCE_HANDLER = None


def get_reference_handler():
    global REFERENCE_HANDLER

    if not REFERENCE_HANDLER:
        embed_handlers = features.get_embed_types()
        link_handlers = features.get_link_types()
        REFERENCE_HANDLER = MultiTypeRichTextReferenceHandler([
            RichTextReferenceHandler(link_handlers, FIND_A_TAG, 'linktype'),
            RichTextReferenceHandler(embed_handlers, FIND_EMBED_TAG, 'embedtype')
        ])
    return REFERENCE_HANDLER



from django.conf import settings
from django.contrib.auth.models import Permission
from django.templatetags.static import static
from django.urls import include, re_path, reverse
from wagtail import hooks
from wagtail.admin.menu import MenuItem

from . import admin_urls


@hooks.register('register_admin_urls')
def register_admin_urls():
    return [
        re_path(r'^wagtail-transfer/', include(admin_urls, namespace='wagtail_transfer_admin')),
    ]


class WagtailTransferMenuItem(MenuItem):
    def is_shown(self, request):
        return all(
            [
                bool(getattr(settings, "WAGTAILTRANSFER_SOURCES", None)),
                request.user.has_perm("wagtail_transfer.wagtailtransfer_can_import"),
            ]
        )


@hooks.register('register_admin_menu_item')
def register_admin_menu_item():
    return WagtailTransferMenuItem('Import', reverse('wagtail_transfer_admin:choose_page'), classnames='icon icon-doc-empty-inverse', order=10000)


@hooks.register("register_permissions")
def register_wagtail_transfer_permission():
    return Permission.objects.filter(
        content_type__app_label="wagtail_transfer",
        codename="wagtailtransfer_can_import",
    )



import uuid

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError

from wagtail_transfer.models import (IDMapping, get_base_model,
                                     get_model_for_path)

# Namespace UUID common to all wagtail-transfer installances, used with uuid5 to generate
# a predictable UUID for any given model-name / PK combination
NAMESPACE = uuid.UUID('418b5168-5a10-11ea-a84b-7831c1c42e66')


class Command(BaseCommand):
    help = "Pre-seed ID mappings used for content transfer"

    def add_arguments(self, parser):
        parser.add_argument('labels', metavar='model_or_app', nargs='+', help="Model (as app_label.model_name) or app name to populate table entries for, e.g. wagtailcore.Page or wagtailcore")
        parser.add_argument('--range', help="Range of IDs to create mappings for (e.g. 1-1000)")

    def handle(self, *args, **options):
        models = []
        for label in options['labels']:
            label = label.lower()
            if '.' in label:
                # interpret as a model
                try:
                    model = get_model_for_path(label)
                except ObjectDoesNotExist:
                    raise CommandError("%r is not recognised as a model name." % label)

                if model != get_base_model(model):
                    raise CommandError(
                        "%r is not a valid model for ID mappings, as it is a subclass using multi-table inheritance." % label
                    )

                models.append(model)
            else:
                # interpret label as an app
                try:
                    app = apps.get_app_config(label)
                except LookupError:
                    raise CommandError("%r is not recognised as an app label." % label)

                for model in app.get_models():
                    if model == get_base_model(model):
                        models.append(model)

        created_count = 0

        for model in models:
            model_name = "%s.%s" % (model._meta.app_label, model._meta.model_name)

            content_type = ContentType.objects.get_for_model(model)
            # find IDs of instances of this model that already exist in the IDMapping table
            mapped_ids = IDMapping.objects.filter(content_type=content_type).values_list('local_id', flat=True)

            # these will be returned as strings, so convert to the pk field's native type
            mapped_ids = [model._meta.pk.to_python(id) for id in mapped_ids]

            # find IDs of instances not in this set
            unmapped_ids = model.objects.exclude(pk__in=mapped_ids).values_list('pk', flat=True)

            # apply ID range filter if passed
            if options['range']:
                min_id, max_id = options['range'].split('-')
                unmapped_ids = unmapped_ids.filter(pk__gte=min_id, pk__lte=max_id)

            # create ID mapping for each of these
            for pk in unmapped_ids:
                _, created = IDMapping.objects.get_or_create(
                    content_type=content_type, local_id=pk,
                    defaults={'uid': uuid.uuid5(NAMESPACE, "%s:%s" % (model_name, pk))}
                )
                if created:
                    created_count += 1

        if options['verbosity'] >= 1:
            self.stdout.write("%d ID mappings created." % created_count)



from collections import OrderedDict

from rest_framework.authentication import SessionAuthentication
from wagtail import VERSION as WAGTAIL_VERSION
from wagtail.admin.navigation import get_explorable_root_page
from wagtail.models import Page, UserPagePermissionsProxy

if WAGTAIL_VERSION >= (5, 1):
    from wagtail.permission_policies.pages import PagePermissionPolicy

from ..wagtail_api_v2.views import PagesAPIViewSet
from .filters import ForExplorerFilter, HasChildrenFilter
from .serializers import AdminPageSerializer


class PagesAdminAPIViewSet(PagesAPIViewSet):
    base_serializer_class = AdminPageSerializer
    authentication_classes = [SessionAuthentication]

    # Add has_children filter
    filter_backends = PagesAPIViewSet.filter_backends + [
        HasChildrenFilter,
    ]

    meta_fields = PagesAPIViewSet.meta_fields + [
        'latest_revision_created_at',
        'status',
        'children',
        'descendants',
        'parent',
        'ancestors',
    ]

    body_fields = PagesAPIViewSet.body_fields + [
        'admin_display_title',
    ]

    listing_default_fields = PagesAPIViewSet.listing_default_fields + [
        'latest_revision_created_at',
        'status',
        'children',
        'admin_display_title',
    ]

    # Allow the parent field to appear on listings
    detail_only_fields = []

    known_query_parameters = PagesAPIViewSet.known_query_parameters.union([
        'has_children'
    ])

    def get_root_page(self):
        """
        Returns the page that is used when the `&child_of=root` filter is used.
        """
        return Page.get_first_root_node()

    def get_base_queryset(self):
        """
        Returns a queryset containing all pages that can be seen by this user.

        This is used as the base for get_queryset and is also used to find the
        parent pages when using the child_of and descendant_of filters as well.
        """
        return Page.objects.all()

    def get_queryset(self):
        queryset = super().get_queryset()

        # Hide root page
        # TODO: Add "include_root" flag
        # .specific() is required to use .get_admin_display_title()
        queryset = queryset.exclude(depth=1).specific()
        return queryset

    def get_type_info(self):
        types = OrderedDict()

        for name, model in self.seen_types.items():
            types[name] = OrderedDict([
                ('verbose_name', model._meta.verbose_name),
                ('verbose_name_plural', model._meta.verbose_name_plural),
            ])

        return types

    def listing_view(self, request):
        response = super().listing_view(request)
        response.data['__types'] = self.get_type_info()
        return response

    def detail_view(self, request, pk):
        response = super().detail_view(request, pk)
        response.data['__types'] = self.get_type_info()
        return response


class PagesForExplorerAdminAPIViewSet(PagesAdminAPIViewSet):
    # Add for_explorer filter
    filter_backends = PagesAPIViewSet.filter_backends + [
        ForExplorerFilter,
    ]

    def get_root_page(self):
        if WAGTAIL_VERSION >= (5, 1):
            return PagePermissionPolicy().explorable_root_instance(self.request.user)
        return get_explorable_root_page(self.request.user)

    def get_base_queryset(self, models=None):
        queryset = super().get_base_queryset(models=models)

        if WAGTAIL_VERSION >= (5, 1):
            permission_policy = PagePermissionPolicy()
            queryset = queryset & permission_policy.explorable_instances(self.request.user)
        else:
            user_perms = UserPagePermissionsProxy(self.request.user)
            queryset = queryset & user_perms.explorable_pages()

        return queryset



from django.urls import re_path

from wagtail_transfer import views

from .vendor.wagtail_api_v2.router import WagtailAPIRouter

chooser_api = WagtailAPIRouter('wagtail_transfer_admin:page_chooser_api')
chooser_api.register_endpoint('pages', views.PageChooserAPIViewSet)

app_name = 'wagtail_transfer_admin'
urlpatterns = [
    re_path(r'^choose/$', views.choose_page, name='choose_page'),
    re_path(r'^import/$', views.do_import, name='import'),
    re_path(r'^api/chooser-local/', (chooser_api.urls[0], 'page_chooser_api', 'page_chooser_api')),
    re_path(r'^api/chooser-proxy/(\w+)/([\w\-/]*)$', views.chooser_api_proxy, name='chooser_api_proxy'),
    re_path(r'^api/check_uid/$', views.check_page_existence_for_uid, name='check_uid'),
]



from functools import partial

from django.core.exceptions import ValidationError
from wagtail.blocks import (ChooserBlock, ListBlock, RichTextBlock,
                            StreamBlock, StructBlock)

from .models import get_base_model
from .richtext import get_reference_handler


def get_references_using_handler(block, stream, references):
    """Gets object references from a streamfield with block object block and value stream, and updates the reference
    set with the result"""
    block_handler = get_block_handler(block)
    references.update(block_handler.get_object_references(stream))
    return stream


def update_ids_using_handler(block, stream, destination_ids_by_source):
    """Updates reference ids from source to destination site for a streamfield block, using its handler"""
    block_handler = get_block_handler(block)
    return block_handler.update_ids(stream, destination_ids_by_source)


def get_object_references(stream_block, stream):
    """Loops over list of dicts formatted StreamField (stream) to find object references. This format is used as opposed
    to the StreamChild object format to prevent ChooserBlocks trying to load nonexistent models with old ids upon to_python
    being called"""
    references = set()
    get_references = partial(get_references_using_handler, references=references)
    stream_block_handler = get_block_handler(stream_block)
    try:
        stream_block_handler.map_over_json(stream, get_references)
    except ValidationError:
        pass
    return references


def update_object_ids(stream_block, stream, destination_ids_by_source):
    """Loops over list-of-dicts formatted StreamField (stream) to update object references. This format is used as opposed
    to the StreamChild object format to prevent ChooserBlocks trying to load nonexistent models with old ids upon to_python
    being called"""
    update_ids = partial(update_ids_using_handler, destination_ids_by_source=destination_ids_by_source)
    stream_block_handler = get_block_handler(stream_block)
    try:
        updated_stream = stream_block_handler.map_over_json(stream, update_ids)
    except ValidationError:
        updated_stream = []
    return updated_stream


class BaseBlockHandler:
    """Base class responsible for finding object references and updating ids for StreamField blocks"""

    empty_value = None

    def __init__(self, block):
        self.block = block

    def get_object_references(self, value):
        """
        Return a set of (model_class, id) pairs for all objects referenced in this block
        """
        return set()

    def update_ids(self, value, destination_ids_by_source):
        """
        Update source site ids to destination site ids
        """
        return value

    def map_over_json(self, stream, func):
        """
        Apply a function, func, to each of the base blocks' values (ie not Struct, List, Stream) of a StreamField in
        list of dicts (imported json) format and return a copy of the rewritten streamfield.
        """
        value = func(self.block, stream)
        if self.block.required and value is None:
            raise ValidationError('This block requires a value')
        return value


class ListBlockHandler(BaseBlockHandler):
    def map_over_json(self, stream, func):
        updated_stream = []
        new_block = self.block.child_block
        new_block_handler = get_block_handler(new_block)
        block_is_in_new_format = getattr(
            self.block,
            "_item_is_in_block_format",
            lambda x: False
        )
        for element in stream:
            try:
                if block_is_in_new_format(element):
                    # We are dealing with new-style ListBlock representation
                    new_value = new_block_handler.map_over_json(element['value'], func)
                    updated_stream.append({'type': element['type'], 'value': new_value, 'id': element['id']})
                else:
                    new_value = new_block_handler.map_over_json(element, func)
                    updated_stream.append(new_value)
            except ValidationError:
                pass
        return updated_stream

    @property
    def empty_value(self):
        return []


class StreamBlockHandler(BaseBlockHandler):
    def map_over_json(self, stream, func):
        updated_stream = []
        for element in stream:
            new_block = self.block.child_blocks.get(element['type'])
            new_block_handler = get_block_handler(new_block)
            new_stream = element['value']
            try:
                new_value = new_block_handler.map_over_json(new_stream, func)
                updated_stream.append({'type': element['type'], 'value': new_value, 'id': element['id']})
            except ValidationError:
                # Omit the block if a required field was left blank due to the import
                pass
        if self.block.required and not updated_stream:
            raise ValidationError('This block requires a value')
        return updated_stream

    @property
    def empty_value(self):
        return []


class StructBlockHandler(BaseBlockHandler):
    remove_if_empty = True

    def map_over_json(self, stream, func):
        updated_stream = {}
        for key in stream:
            new_block = self.block.child_blocks.get(key)
            new_block_handler = get_block_handler(new_block)
            new_stream = stream[key]
            try:
                new_value = new_block_handler.map_over_json(new_stream, func)
            except ValidationError:
                if new_block.required:
                    raise ValidationError('This block requires a value for {}'.format(new_block))
                else:
                    # If the new block isn't required, just set it to the empty value
                    new_value = new_block_handler.empty_value
            updated_stream[key] = new_value
        return updated_stream


class RichTextBlockHandler(BaseBlockHandler):
    def get_object_references(self, value):
        return get_reference_handler().get_objects(value)

    def update_ids(self, value, destination_ids_by_source):
        value = get_reference_handler().update_ids(value, destination_ids_by_source)
        return value


class ChooserBlockHandler(BaseBlockHandler):
    def get_object_references(self, value):
        if value:
            return {(get_base_model(self.block.model_class), value)}
        return set()

    def update_ids(self, value, destination_ids_by_source):
        value = destination_ids_by_source.get((get_base_model(self.block.model_class), value))
        return value


def get_block_handler(block):
    # find the handler class for the most specific class in the block's inheritance tree
    for block_class in type(block).__mro__:
        if block_class in HANDLERS_BY_BLOCK_CLASS:
            handler_class = HANDLERS_BY_BLOCK_CLASS[block_class]
            return handler_class(block)
    return BaseBlockHandler(block)


HANDLERS_BY_BLOCK_CLASS = {
    RichTextBlock: RichTextBlockHandler,
    ChooserBlock: ChooserBlockHandler,
    ListBlock: ListBlockHandler,
    StreamBlock: StreamBlockHandler,
    StructBlock: StructBlockHandler,
}
